/***********************************************************************/
/*                                                                     */
/*  FenForm - Document Printer                                         */
/*  Vn 1.4                                                             */
/*  Author - Mike Eggleston                                            */
/*  Copyright 2005 Fenland Software Ltd                                */
/*  Date 05/01/2005                                                    */
/*                                                                     */
/***********************************************************************/

#define  PROGNAME	"FenForm"
#define  VERSION	"Vn 1.4"
#define  COPYRIGHT	"Copyright 2005 Fenland Software Ltd"

#include "fenform.h"
#include "barcode.h"

/******************************/
/*  Global Variables          */
/******************************/

char config_file[FULLPATH + 1];
struct FF_config cfg;
struct FF_layout frm;
struct FF_status sts;
struct FF_barcode bc_parm;

struct passwd *pswd;
char *date_string = NULL;

int verbose;

FILE *ifile, *ofile;

/******************************/
/*  Forward definitions       */
/******************************/

void set_colour(void);
void set_new_font(void);


/******************************/
/*  Utility Functions         */
/******************************/

/* Remove all 'new-line' characters from the end of a string */
void remove_newline(char *textline)
{
  while(textline[strlen(textline)-1] == '\n')
    textline[strlen(textline)-1] = '\0';
}

/* Send an e-mail to the administrator if invalid parameter found */
int invalid_param(char *parm, char origin)
{
char message[EMAILLENGTH];

  syslog(LOG_ERR, "Invalid configuration parameter: %s", parm);

  if(*cfg.admin) {
    syslog(LOG_ERR, "email sent to %s", cfg.admin);

    if(origin == 'C') {
      sprintf(message, "echo \"Subject: FenForm Parameter Error\n\
Content-Type:text/plain; charset=\"us-ascii\"\n\
Content-Transfer-Encoding: 7bit\n\nThe FenForm configuration file (%s)\n\
contains the invalid parameter \"%s\".\n\nThis parameter will be \
ignored.\n\n\n\" | %s %s", config_file, parm, cfg.mailer, cfg.admin);
    }
    else
    {
      sprintf(message, "echo \"Subject: FenForm Parameter Error\n\
Content-Type:text/plain; charset=\"us-ascii\"\n\
Content-Transfer-Encoding: 7bit\n\nThe FenForm form file (%s)\n\
contains the invalid parameter \"%s\".\n\nThis parameter will be \
ignored.\n\n\n\" | %s %s", sts.curr_form, parm, cfg.mailer, cfg.admin);
    }
  
    return(system(message));
  }
  else
    return(0);
}

int extract_parm(char *full_parm, char *parm, char *val)
{
char parm_str[LINE_LN + 1];
char p_val[LINE_LN + 1];
char c_val[LINE_LN + 1];

  strncpy(parm_str, full_parm, LINE_LN);
  memset(p_val,'\0',LINE_LN + 1);
  memset(c_val,'\0',LINE_LN + 1);

  memcpy(p_val,parm_str,strcspn(parm_str,"="));
  if(strlen(p_val) < strlen(parm_str))
  {
    strcpy(c_val,strchr(parm_str,'=')+1);
    remove_newline(c_val);
  }
  else
    *c_val = 0;

  remove_newline(p_val);
  while(p_val[strlen(p_val)-1] == ' ')
    p_val[strlen(p_val)-1] = '\0';

  strcpy(parm,p_val);
  strcpy(val, c_val);

return 0;
}

int get_config(char *cfg_file_name, int mode)
{
FILE *cfg_file;
char parm_str[LINE_LN + 1];
char parm[LINE_LN + 1];
char c_val[LINE_LN + 1];
int ret = 0;

  /* Clear the configuration structure */
  memset(&cfg, '\0', sizeof(struct FF_config));

  /* Set up configuration defaults.  */
  /* These will be overridden by a configuration file  */

  strcpy(cfg.forms_dir, _FORMS);
  strcpy(cfg.std_font, _STDFONT);
  strcpy(cfg.bold_font, _STDBOLD);
  strcpy(cfg.dblstrike_font, _STDDBLSTRIKE);
  strcpy(cfg.italic_font, _STDITALIC);
  strcpy(cfg.bold_italic_font, _STDBOLDITALIC);
  strcpy(cfg.dblstrike_italic_font, _STDDBLSTRIKEITALIC);
  strcpy(cfg.nlq_font, _NLQFONT);
  strcpy(cfg.bold_nlq_font, _NLQBOLD);
  strcpy(cfg.dblstrike_nlq_font, _NLQDBLSTRIKE);
  strcpy(cfg.nlq_italic_font, _NLQITALIC);
  strcpy(cfg.bold_nlq_italic_font, _NLQBOLDITALIC);
  strcpy(cfg.dblstrike_nlq_italic_font, _NLQDBLSTRIKEITALIC);
  cfg.h_offset = _HOFFSET;
  cfg.v_offset = _VOFFSET;
  cfg.full_colour = FF_OFF;
  strcpy(cfg.mailer, _MAILER);
  cfg.line_width = BASE_LINEWIDTH;

  if(mode)
    fprintf(stderr,"Validating Configuration file:  %s\n", cfg_file_name);

  if(cfg_file = fopen(cfg_file_name, "r"))
  {
    syslog(LOG_INFO,"Reading configuration file %s\n", cfg_file_name);
    while(fgets(parm_str, LINE_LN, cfg_file))
    {
     
      if((parm_str[0] == '#') && (parm_str[strlen(parm_str)-1] != '\n'))
      {
        /* Continue reading to end of line  */
        while(fgetc(cfg_file) != '\n'); 

        continue;
      }

      if(parm_str[0] != '#' && parm_str[0] != ' ')
      {
        extract_parm(parm_str, parm, c_val);
        if(parm[0] == 0)
          continue;

        if(strcasecmp(parm,"FORMDIR") == '\0')
        {
          strcpy(cfg.forms_dir, strcat(c_val, "/"));
          continue;
        }

        if(strcasecmp(parm,"STDFONT") == '\0')
        {
          strcpy(cfg.std_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"BOLDFONT") == '\0')
        {
          strcpy(cfg.bold_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"DBLSTRIKEFONT") == '\0')
        {
          strcpy(cfg.dblstrike_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"ITALICS") == '\0')
        {
          strcpy(cfg.italic_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"BOLDITALICS") == '\0')
        {
          strcpy(cfg.bold_italic_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"DBLSTRIKEITALICS") == '\0')
        {
          strcpy(cfg.dblstrike_italic_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"NLQFONT") == '\0')
        {
          strcpy(cfg.nlq_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"BOLDNLQFONT") == '\0')
        {
          strcpy(cfg.bold_nlq_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"DBLSTRIKENLQFONT") == '\0')
        {
          strcpy(cfg.dblstrike_nlq_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"NLQITALICS") == '\0')
        {
          strcpy(cfg.nlq_italic_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"BOLDNLQITALICS") == '\0')
        {
          strcpy(cfg.bold_nlq_italic_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"DBLSTRIKENLQITALICS") == '\0')
        {
          strcpy(cfg.dblstrike_nlq_italic_font, c_val);
          continue;
        }

        if(strcasecmp(parm,"COLOUR") == '\0')
        {
          if(strcasecmp(c_val, "yes"))
            cfg.full_colour = FF_ON;
          else
            cfg.full_colour = FF_OFF;
          continue;
        }

        if(strcasecmp(parm,"MAILER") == '\0')
        {
          strcpy(cfg.mailer, c_val);
          continue;
        }

        if(strcasecmp(parm,"ADMINISTRATOR") == '\0')
        {
          strcpy(cfg.admin, c_val);
          continue;
        }

        /* If we get here, the parameter name is invalid */
       ret = 1;

       if(mode)
         fprintf(stderr,"Invalid parameter: %s\n", parm);
       else
         invalid_param(parm, 'C');
 
      }
    }
  fclose(cfg_file);
  }
  else
  {
    if(strcasecmp(cfg_file_name, _CFG_FILE))
    {
      syslog(LOG_INFO,"Configuration file (%s) not found\n", cfg_file_name);
      ret = -1;
      if(mode)
        fprintf(stderr,"Configuration file not found.\n");

    }
    else
    {
      if(verbose)
        syslog(LOG_INFO,"No configuration file - using defaults");
    }
  }

return(ret);
}

/* Initialize the "printer".  This sets the state of Fenform */
/* to that of a printer immediately after power-up.          */
void init()
{
int n, tabpos = 0;

  strcpy(sts.curr_font, cfg.std_font);
  frm.line_space = LPI6;
  frm.char_width = CPI10;
  frm.l_margin = 0;
  frm.r_margin = 0;
  sts.italics = FF_OFF;
  sts.bold = FF_OFF;
  sts.dblstrike = FF_OFF;
  sts.nlq = FF_OFF;
  sts.underline = FF_OFF;
  sts.h_tab = 0;
  sts.v_tab = 0;
  sts.cur_line = frm.form_length - cfg.v_offset - frm.line_space;
  sts.start_col = cfg.h_offset + frm.l_margin;
  sts.cur_col = sts.start_col;
  sts.text_mode = FF_OFF;
  sts.red = 0;
  sts.green = 0;
  sts.blue = 0;

  /* Set default tabs. */
  for(n=0; n< MAX_H_TABS; n++)
  {
    tabpos += DFT_HTAB;
    frm.htabs[n] = tabpos * frm.char_width;
  }
}

 
void default_form(void)
{
  /* Page parameters */
  strcpy(frm.form_name,"A4P");
  frm.orientation = 'P';
  frm.form_length = DFT_PAPER_L;
  frm.form_width = DFT_PAPER_W;
}

int get_form_parms(FILE *formp)
{
char parm_str[LINE_LN + 1];
char parm[LINE_LN + 1];
char c_val[LINE_LN + 1];

  /* Default Parameters if a new print file */
  if(sts.new_page == FF_OFF)
    default_form();

  fgets(parm_str,LINE_LN,formp);
  if(strcasecmp(parm_str,"*PARMS\n"))
  {
    fseek(formp,0L,SEEK_SET);
    return 0;
  }

  while(fgets(parm_str, LINE_LN, formp) && (strcasecmp(parm_str,"**END\n")))
  {
    extract_parm(parm_str, parm, c_val);

    if(strcasecmp(parm,"NAME") == '\0')
    {
      strcpy(frm.form_name,c_val);
      continue;
    }

    if(strcasecmp(parm,"LENGTH") == '\0')
    {
      frm.form_length = atof(c_val);
      continue;
    }

    if(strcasecmp(parm,"WIDTH") == '\0')
    {
      frm.form_width = atof(c_val);
      continue;
    }

    if(strcasecmp(parm,"LPI") == '\0')
    {
      frm.line_space = 72/atof(c_val);
      continue;
    }

    if(strcasecmp(parm,"CPI") == '\0')
    {
      frm.char_width = 72/atof(c_val);
      continue;
    }

    if(strcasecmp(parm,"HOFFSET") == '\0')
    {
      cfg.h_offset = abs(atof(c_val));
      continue;
    }

    if(strcasecmp(parm,"VOFFSET") == '\0')
    {
      cfg.v_offset = atof(c_val);
      continue;
    }

    /* If we get here, the parameter name is invalid */
    invalid_param(parm, 'F');

  }

  if(( frm.r_margin == 0) || (frm.r_margin > frm.form_width) )
    frm.r_margin = frm.form_width;

return 0;
}

void write_file_header()
{
  /* Write header to Output */
  fprintf(ofile, "%s\n", "%!PS-Adobe-3.0");

  if(frm.form_width > frm.form_length)
  {
    frm.orientation = 'L';
    fprintf(ofile, "%%%%HiResBoundingBox: %0.2f %0.2f %0.2f %0.2f\n", 
	0.00, 0.00, frm.form_length, frm.form_width);
  }
  else
  {
    frm.orientation = 'P';
    fprintf(ofile, "%%%%HiResBoundingBox: %0.2f %0.2f %0.2f %0.2f\n", 
	0.00, 0.00, frm.form_width, frm.form_length);
  }

  fprintf(ofile, "%%%%Title: %s\n", "FenForm Output");
  fprintf(ofile, "%%%%For: %s\n", pswd->pw_gecos);
  fprintf(ofile, "%%%%Creator: %s %s\n", PROGNAME, VERSION);
  fprintf(ofile, "%%%%CreationDate: %s\n", date_string);
  fprintf(ofile, "%%%%Orientation: %s\n", frm.orientation == 'L' ?
                                               "Landscape" : "Portrait");
  fprintf(ofile, "%%%%Pages: (atend)\n");
  fprintf(ofile, "%%%%PageOrder: Ascend\n");

  if(frm.orientation == 'L')
    fprintf(ofile, "%%%%DocumentMedia: %s %0.2f %0.2f 0 () ()\n",
        frm.form_name, frm.form_length, frm.form_width);
  else
    fprintf(ofile, "%%%%DocumentMedia: %s %0.2f %0.2f 0 () ()\n",
        frm.form_name, frm.form_width, frm.form_length);

  fprintf(ofile, "%%%%DocumentNeededResources: (atend)\n");
  fprintf(ofile, "%%%%LanguageLevel: 2\n");


  fprintf(ofile, "%%%%EndComments\n\n");

  /* Procedure Definitions. */
  fprintf(ofile, "%%%%BeginProcSet: fenformset\n");
  fprintf(ofile, "   /selectfont where { %%ifelse\n     pop\n");
  fprintf(ofile, "   }{ %%else\n     selectfont { %%def\n");
  fprintf(ofile, "       exch findfont exch scalefont setfont\n");
  fprintf(ofile, "   }} bind def\n");

  fprintf(ofile, "/Fnt /selectfont load def\n");
  fprintf(ofile, "/ms {moveto show} bind def\n");
  fprintf(ofile, "/ls {newpath moveto lineto stroke} bind def\n");
  fprintf(ofile, "/rs {rectstroke} bind def\n");
  fprintf(ofile, "/wm {gsave setgray moveto rotate show grestore} bind def\n\n");
  fprintf(ofile, "%%%%EndProcSet\n");

  fprintf(ofile, "%%%%EndProlog\n\n");

}

void write_file_end()
{
  /* Write trailer to Output */
  fprintf(ofile, "%%%%Trailer\n");
  fprintf(ofile, "%%%%Pages: %d\n", sts.page);
  fprintf(ofile, "%%%%DocumentNeededResources: font %s\n%% + font %s\n\
%% + font %s\n%% + font %s\n%% + font %s\n%% + font %s\n%% + font %s\n\
%% + font %s\n", cfg.std_font, cfg.bold_font, cfg.italic_font, 
cfg.bold_italic_font, cfg.nlq_font, cfg.bold_nlq_font, 
cfg.nlq_italic_font, cfg.bold_nlq_italic_font);

  fprintf(ofile, "%%%%EOF\n");
}

int read_form()
{
FILE *FF_form;
int rv;
char str[LINE_LN];

  /* Open form file if either specified or default form required */
  FF_form = 0;
  if(sts.curr_form[0])
  {
    if(!(FF_form = fopen(sts.curr_form, "r")))
    {
      syslog(LOG_ERR, "OPEN failed - %s\n",sts.curr_form);
      return -1;  /* Exit if specified form not found. */
    }
    else
    {
      if(verbose)
        syslog(LOG_INFO, "Form file %s opened\n",sts.curr_form);
    }
  }

  /* Increment the page count */
  sts.page++;
  
  if(FF_form)
  {
    if(!get_form_parms(FF_form))
    {
      /* Initialize Default Parameters if a new print file */
      if(sts.new_page == FF_OFF)
      {
        write_file_header();
        sts.page = 1;
      }

      if(frm.orientation == 'L')
        fprintf(ofile,"90 rotate\n0 -%d translate\n\n",frm.form_length);

      fprintf(ofile,"%%%%Page: (%d) %d\n",sts.page, sts.page);
      fprintf(ofile,"%% Start of Form Specification\n");
      while(fgets(str, LINE_LN, FF_form))
        fprintf(ofile, "%s", str);
    }
    fclose(FF_form);
    fprintf(ofile,"\n%% End of Form Specification\n\n");
  }
  else
  {
    /* Initialize Default Parameters if a new print file */
    if(sts.new_page == FF_OFF)
    {
      default_form();
      write_file_header();
      sts.page = 1;
    }
    fprintf(ofile,"\n%% No Form Specified\n\n");
  }

  /* These are re-defined here in case */
  /* the form file has defined them.   */
  fprintf(ofile, "/ms {moveto show} bind def\n");
  fprintf(ofile, "/ls {newpath moveto lineto stroke} bind def\n");
  fprintf(ofile, "/rs {rectstroke} bind def\n");

  set_colour();
  fprintf(ofile,"%0.2f setlinewidth\n", cfg.line_width);
  sts.font_set = FF_OFF;

return 0;
}

int start_page()
{
  /* Set ready for top line */
  if(read_form())
    return -1;
  else
  {
    sts.h_tab = 0;
    sts.v_tab = 0;
    sts.cur_line = frm.form_length - cfg.v_offset - frm.line_space;
    sts.start_col = cfg.h_offset + frm.l_margin;
    sts.cur_col = sts.start_col;
    sts.text_mode = FF_OFF;
    sts.font_set = FF_OFF;
    sts.new_page = FF_OFF;
    return 0;
  }
}

/* Read numeric (no_parms) parameters from file.                  */  
/* if no_parms = 0, keep reading until NULL or '\' or (max) read  */
int read_parameters(int *parms, int no_parms, int max)
{
unsigned int value;
int x, y;

  if(no_parms)
    max = no_parms;

  for(y=0; y<max; y++)
  {
    x = fgetc(ifile);
    if(no_parms == 0 && x == '\0')
      break;
    parms[y] = x;
  }
  return y;  /* Return the number of parameters read */
}


/******************************/
/*  Print Control Functions   */
/******************************/

void set_new_font(void)
{
  fprintf(ofile,"\n/%s %0.2f selectfont\n",
      sts.curr_font, 120*frm.char_width/72);
  sts.font_set = FF_ON;
}

void set_colour(void)
{
float colour;

  colour = sts.red + sts.green + sts.blue;

  if(cfg.full_colour)
    fprintf(ofile, "%% Full Colour Mode\n");

  if(cfg.full_colour && (colour > 0.0))
    fprintf(ofile, "%0.2f %0.2f %0.2f setrgbcolor\n",
        sts.red, sts.green, sts.blue);
  else
    fprintf(ofile, "0.0 setgray\n");
}

void set_barcode(void)
{
int rv;
int param[3];

  bc_parm.flags = BARCODE_OUT_PS | BARCODE_OUT_NOHEADERS | BARCODE_NO_ASCII;

  if(read_parameters(param, 3, 0) == 3)
  {
    bc_parm.flags |= (unsigned int)param[0];
    bc_parm.height = (unsigned int)param[1];
    bc_parm.width = (unsigned int)param[2];
  }
}

void print_barcode(void)
{
char *text;
int x, y, text_length, err_flg = 0;

  x = (int)sts.cur_col;
  y = (int)sts.cur_line - BARCODE_DEFAULT_MARGIN;

  fprintf(ofile, "\n\n");

  if(read_parameters(&text_length, 1, 1))
  {
    text = malloc(((int)text_length + 1) * sizeof(char));
    fgets(text, text_length + 1, ifile);

    if((2 * BARCODE_DEFAULT_MARGIN + bc_parm.width + x) > frm.r_margin)
      err_flg = -9;
    else
      err_flg = Barcode_Encode_and_Print(text, ofile, bc_parm.width, 
bc_parm.height, x, y, bc_parm.flags);

    if(err_flg < 0)
    {
      /* Invalid Barcode */
      x += BARCODE_DEFAULT_MARGIN;
      y += BARCODE_DEFAULT_MARGIN;
      
      fprintf(ofile, "1 setlinewidth\n");
      fprintf(ofile, "%d %d %d %d rs\n", x, y, bc_parm.width, bc_parm.height);
      fprintf(ofile, "%d %d %d %d ls\n", x, y, x + bc_parm.width, y + bc_parm.height);
      fprintf(ofile, "%d %d %d %d ls\n", x, y + bc_parm.height, x + bc_parm.width, y);
    }

    fprintf(ofile, "%0.2f setlinewidth\n", cfg.line_width);

    free(text);
  }

  sts.font_set = FF_OFF;
  sts.cur_col += (2 * BARCODE_DEFAULT_MARGIN + bc_parm.width);

}

void set_page_length(void)
{
int length;

  if(read_parameters(&length, 1, 1) == 1)
    frm.form_length = (float)(length * LPI6);

  /* If the length returned is zero, read the next */
  /* parameter to get the page length in inches    */
  if(length == 0)
  {
    if(read_parameters(&length, 1, 1) == 1)
      frm.form_length = (float)(length * 72);  /* 'points' per inch */
  }

  if(sts.cur_line > (frm.form_length - cfg.v_offset - frm.line_space))
    sts.cur_line = frm.form_length - cfg.v_offset - frm.line_space;

}

int esc_error(int action)
{
  syslog(LOG_ERR, "Invalid Escape Code (%c) found\n", action);
return 0;
}

void print_underline(void)
{
float line;

  line = sts.cur_line - sts.line_offset;

  fprintf(ofile,"%0.2f %0.2f %0.2f %0.2f ls\n",
      sts.ul_start, line - 2, sts.ul_end, line - 2);
  sts.ul_start = sts.cur_col;
  sts.ul_end = 0;
}

void new_page(void)
{
  if(sts.ul_end > sts.ul_start)  /* Finish current page */
    print_underline();
  if(sts.text_mode == FF_ON)
  {
    fprintf(ofile,") %0.2f %0.2f ms\n", sts.start_col,sts.cur_line);
    sts.text_mode = FF_OFF;
  }
  fprintf(ofile,"showpage\n\n");
  sts.new_page = FF_ON;
}

void back_space(void)
{
  if(sts.cur_col > frm.char_width)
  {
    sts.start_col = sts.cur_col - frm.char_width;
    sts.cur_col = sts.start_col;
  }
}

void line_feed(void)
{
  sts.h_tab = 0;
  sts.start_col = cfg.h_offset + frm.l_margin;
  sts.cur_col = sts.start_col;

  if(sts.ul_end > sts.ul_start)
    print_underline();

  if(sts.cur_line >= frm.line_space)
    sts.cur_line -= frm.line_space;
  else
    sts.cur_line = 0;
}

void horizontal_tab(void)
{
  if(sts.cur_col < frm.form_width)
  {
    while(frm.htabs[sts.h_tab] <= sts.cur_col)
      sts.h_tab++;
    sts.cur_col = frm.htabs[sts.h_tab];
    sts.start_col = sts.cur_col;

    if(sts.ul_end > sts.ul_start)
      print_underline();

    sts.h_tab++;
  }
}

void vertical_tab(void)
{
  if(sts.ul_end > sts.ul_start)
    print_underline();

  if(sts.cur_line < (frm.vtabs[sts.v_tab] - cfg.v_offset))
    new_page();
  else
  {
    sts.cur_line = frm.vtabs[sts.v_tab] - cfg.v_offset;
    sts.v_tab++;
    sts.h_tab = 0;
    sts.start_col = cfg.h_offset + frm.l_margin;
    sts.cur_col = sts.start_col;
  }
}

void set_italics(void)
{
  sts.italics = FF_ON;
  strcpy(sts.curr_font, cfg.italic_font);
  if(sts.bold)
  strcpy(sts.curr_font, cfg.bold_italic_font);
  if(sts.dblstrike)
  strcpy(sts.curr_font, cfg.dblstrike_italic_font);
  if(sts.nlq) 
    strcpy(sts.curr_font, cfg.nlq_italic_font);
  if(sts.nlq && sts.bold) 
    strcpy(sts.curr_font, cfg.bold_nlq_italic_font);
  sts.font_set = FF_OFF;
}

void cancel_italics(void)
{
  sts.italics = FF_OFF;
  strcpy(sts.curr_font, cfg.std_font);
  if(sts.bold)
    strcpy(sts.curr_font, cfg.bold_font);
  if(sts.dblstrike)
    strcpy(sts.curr_font, cfg.dblstrike_font);
  if(sts.nlq) 
    strcpy(sts.curr_font, cfg.nlq_font);
  if(sts.nlq && sts.bold) 
    strcpy(sts.curr_font, cfg.bold_nlq_font);
  sts.font_set = FF_OFF;
}

void set_bold_font(void)
{
  sts.bold = FF_ON;
  strcpy(sts.curr_font, cfg.bold_font);
  if(sts.italics) 
    strcpy(sts.curr_font, cfg.bold_italic_font);
  if(sts.nlq) 
    strcpy(sts.curr_font, cfg.bold_nlq_font);
  if(sts.nlq && sts.italics) 
    strcpy(sts.curr_font, cfg.bold_nlq_italic_font);
  sts.font_set = FF_OFF;
}

void cancel_bold_font(void)
{
  sts.bold = FF_OFF;
  strcpy(sts.curr_font, cfg.std_font);
  if(sts.dblstrike)
    strcpy(sts.curr_font, cfg.dblstrike_font);
  if(sts.italics) 
    strcpy(sts.curr_font, cfg.italic_font);
  if(sts.nlq) 
    strcpy(sts.curr_font, cfg.nlq_font);
  if(sts.nlq && sts.italics) 
    strcpy(sts.curr_font, cfg.nlq_italic_font);
  sts.font_set = FF_OFF;
}

void set_dblstrike_font(void)
{
  sts.dblstrike = FF_ON;
  strcpy(sts.curr_font, cfg.dblstrike_font);
  if(sts.italics) 
    strcpy(sts.curr_font, cfg.dblstrike_italic_font);
  if(sts.nlq) 
    strcpy(sts.curr_font, cfg.dblstrike_nlq_font);
  if(sts.nlq && sts.italics) 
    strcpy(sts.curr_font, cfg.dblstrike_nlq_italic_font);
  sts.font_set = FF_OFF;
}

void cancel_dblstrike_font(void)
{
  sts.dblstrike = FF_OFF;
  strcpy(sts.curr_font, cfg.std_font);
  if(sts.bold)
    strcpy(sts.curr_font, cfg.bold_font);
  if(sts.italics) 
    strcpy(sts.curr_font, cfg.italic_font);
  if(sts.nlq) 
    strcpy(sts.curr_font, cfg.nlq_font);
  if(sts.nlq && sts.italics) 
    strcpy(sts.curr_font, cfg.nlq_italic_font);
  sts.font_set = FF_OFF;
}

void set_nlq_font(void)
{
  sts.nlq = FF_ON;
  strcpy(sts.curr_font, cfg.nlq_font);
  if(sts.italics) 
    strcpy(sts.curr_font, cfg.nlq_italic_font);
  if(sts.bold) 
    strcpy(sts.curr_font, cfg.bold_nlq_font);
  if(sts.dblstrike)
    strcpy(sts.curr_font, cfg.dblstrike_font);
  if(sts.bold && sts.italics) 
    strcpy(sts.curr_font, cfg.bold_nlq_italic_font);
  if(sts.dblstrike && sts.italics) 
    strcpy(sts.curr_font, cfg.dblstrike_nlq_italic_font);
  sts.font_set = FF_OFF;
}

void cancel_nlq_font(void)
{
  sts.nlq = FF_OFF;
  strcpy(sts.curr_font, cfg.std_font);
  if(sts.italics) 
    strcpy(sts.curr_font, cfg.italic_font);
  if(sts.bold)
    strcpy(sts.curr_font, cfg.bold_font);
  if(sts.dblstrike)
    strcpy(sts.curr_font, cfg.dblstrike_font);
  if(sts.bold && sts.italics) 
    strcpy(sts.curr_font, cfg.bold_italic_font);
  if(sts.dblstrike && sts.italics) 
    strcpy(sts.curr_font, cfg.dblstrike_italic_font);
  sts.font_set = FF_OFF;
}

void set_char_width(float width)
{
  sts.prev_char_width = frm.char_width;
  frm.char_width = width;
  if(sts.text_mode == FF_ON)
  {
    fprintf(ofile,") %0.2f %0.2f ms\n",sts.start_col,sts.cur_line);
    sts.text_mode = FF_OFF;
  }
  sts.font_set = FF_OFF;
}

void set_sub_super_script(float sub_super)
{
  if(sts.text_mode == FF_ON)
  {
    fprintf(ofile,") %0.2f %0.2f ms\n",sts.start_col,sts.cur_line);
    sts.text_mode = FF_OFF;
  }
  frm.char_width /= 1.5;
  sts.font_set = FF_OFF;

  if(sub_super == 0)
  {
    sts.sub_superscript = SUPERSCRIPT;
    sts.line_offset = frm.line_space / 3;
  }
  else
  {
    sts.sub_superscript = SUBSCRIPT;
    sts.line_offset = (-1 * frm.line_space / 9);
  }

  sts.cur_line += sts.line_offset;
}

void cancel_sub_super_script(void)
{
  if(sts.text_mode == FF_ON)
  {
    fprintf(ofile,") %0.2f %0.2f ms\n",sts.start_col,sts.cur_line);
    sts.text_mode = FF_OFF;
  }
  frm.char_width *= 1.5;
  sts.font_set = FF_OFF;

  sts.cur_line -= sts.line_offset;
  sts.line_offset = 0;
  sts.sub_superscript = FF_OFF;
}

void set_colour_value(int colour)
{
int rgb[3];

  if(sts.text_mode == FF_ON)
  {
    fprintf(ofile,") %0.2f %0.2f ms\n",sts.start_col,sts.cur_line);
    sts.text_mode = FF_OFF;
  }

  /* Codes 0 - 7 give coarse colour control */
  switch(colour)
  {
    case 0: /* Black */
            sts.red = 0.0;
            sts.green = 0.0;
            sts.blue = 0.0;
            break;
  
    case 1: /* Red */
            sts.red = (float) ((_COLOUR_RED / 0x10000) / 255.0);
            sts.green = (float) (((_COLOUR_RED / 0x100) % 0x100) / 255.0);
            sts.blue = (float) ((_COLOUR_RED % 0x100) / 255.0);
            break;

    case 2: /* Blue */
            sts.red = (float) ((_COLOUR_BLUE / 0x10000) / 255.0);
            sts.green = (float) (((_COLOUR_BLUE / 0x100) % 0x100) / 255.0);
            sts.blue = (float) ((_COLOUR_BLUE % 0x100) / 255.0);
            break;
  
    case 3: /* Violet */
            sts.red = (float) ((_COLOUR_VIOLET / 0x10000) / 255.0);
            sts.green = (float) (((_COLOUR_VIOLET / 0x100) % 0x100) / 255.0);
            sts.blue = (float) ((_COLOUR_VIOLET % 0x100) / 255.0);
            break;
  
    case 4: /* Yellow */
            sts.red = (float) ((_COLOUR_YELLOW / 0x10000) / 255.0);
            sts.green = (float) (((_COLOUR_YELLOW / 0x100) % 0x100) / 255.0);
            sts.blue = (float) ((_COLOUR_YELLOW % 0x100) / 255.0);
            break;
  
    case 5: /* Orange */
            sts.red = (float) ((_COLOUR_ORANGE / 0x10000) / 255.0);
            sts.green = (float) (((_COLOUR_ORANGE / 0x100) % 0x100) / 255.0);
            sts.blue = (float) ((_COLOUR_ORANGE % 0x100) / 255.0);
            break;
  
    case 6: /* Green */
            sts.red = (float) ((_COLOUR_GREEN / 0x10000) / 255.0);
            sts.green = (float) (((_COLOUR_GREEN / 0x100) % 0x100) / 255.0);
            sts.blue = (float) ((_COLOUR_GREEN % 0x100) / 255.0);
            break;
  
    case 255: /* Fine colour control */
              read_parameters(rgb, 3, 3);
              sts.red = (float)rgb[0]/255.0;
              sts.green = (float)rgb[1]/255.0;
              sts.blue = (float)rgb[2]/255.0;
              break;

    default: sts.red = 0;
             sts.green = 0;
             sts.blue = 0;
  
  }
  set_colour();
}
    
/* Process Control Codes */
void ctrl_code(int action)
{
unsigned int rv, cnt;
float *param;

  if((sts.text_mode == FF_ON) && (sts.cur_col > sts.start_col))
  {
    fprintf(ofile,") %0.2f %0.2f ms\n",sts.start_col, sts.cur_line);
    sts.text_mode = FF_OFF;
  }

  if(sts.new_page == FF_ON && action != '@')
    start_page();

  switch(action)
  {
    case '\x08': back_space();
                 break;
    case '\x09': horizontal_tab();
                 break;
    case '\x0a': line_feed();
                 break;
    case '\x0b': vertical_tab();
                 break;
    case '\x0c': new_page();
                 break;
    case '\x0d':  /* Carriage Return  - ** IGNORE ** */
                 break;

    case '\x0f':  /* Select Condensed */
                 if(frm.char_width == (float)CPI10)
                   set_char_width(CPI17);
                 else
                   set_char_width(CPI20);
                 break;

     case '\x12':  /* Cancel Condensed */
                 set_char_width(sts.prev_char_width);
                 break;
  }
}

/* Process Escape sequences */
void escape(int action)
{
unsigned int rv, cnt;
int *param;

  if((sts.text_mode == FF_ON) && (sts.cur_col > sts.start_col))
  {
    fprintf(ofile,") %0.2f %0.2f ms\n",sts.start_col, sts.cur_line);
    sts.text_mode = FF_OFF;
  }

  if(sts.new_page == FF_ON && action != '@')
    start_page();

  switch(action)
  {
    case '\x10':  /* Print Barcode */
              param = malloc(sizeof(int));
              rv = read_parameters(param, 1, 1);
              switch(*param)
              {
                case 'A':  set_barcode();
                           break;
                case 'B':  print_barcode();
                           break;
              }
              free(param);
              break;

    case '!':   /* Select Print Mode - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    case '-': /* Select or Cancel underline */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              if(*param)
              {
                if(!sts.underline)
                {
                  sts.ul_start = sts.cur_col;
                  sts.underline = FF_ON;
                }
              }
              else
                sts.underline = FF_OFF;
              free(param);
         break;

    case '0': frm.line_space = LPI8;  /* Set 8 LPI */
              break;

    case '1': frm.line_space = LPI7_72;  /* Set 7/72 LPI */
              break;

    case '2':
         frm.line_space = LPI6;  /* Set 6 LPI */
         break;

    case '4': set_italics();
              break;
              
    case '5': cancel_italics();
              break;

    case '6':  /* Select Printable Code Area Expansion - ** IGNORE ** */
              break;

    case '7':  /* Cancel Printable Code Area Expansion - ** IGNORE ** */
              break;

    case '8':  /* Disable paper out detection - ** IGNORE ** */
              break;

    case '9':  /* Enable paper out detection - ** IGNORE ** */
              break;

    case '<':  /* Home Head - ** IGNORE ** */
              break;

    case '=':  /* Unset 8th bit (=0) - ** IGNORE ** */
              break;

    case '>':  /* Set 8th bit (=1) - ** IGNORE ** */
              break;

    case '@': /* Reset Printer */
              init();
              break;

    case 'B': /* Set Vertical Tabs */
        param = malloc(sizeof(int) * MAX_V_TABS);
        rv = read_parameters(param, 0, MAX_V_TABS);
        for(cnt = 0; cnt < rv; cnt++)
          frm.vtabs[cnt] = (float)(frm.form_length - param[cnt] * frm.line_space);
        free(param);
        break;

    case 'C': set_page_length();
              break;

    case 'D': /* Set Horizontal Tabs */
        param = malloc(sizeof(int) * MAX_H_TABS);
        rv = read_parameters(param, 0, MAX_H_TABS);
        for(cnt = 0; cnt < rv; cnt++)
          frm.htabs[cnt] = (float)(param[cnt] * frm.char_width);
        for(cnt = rv; cnt < MAX_H_TABS; cnt++)
          frm.htabs[cnt] = frm.htabs[cnt-1] + (DFT_HTAB * frm.char_width);
        free(param);
        break;

    case 'E': set_bold_font();
              break;

    case 'F': cancel_bold_font();
              break;

    case 'G': set_dblstrike_font();
              break;

    case 'H': cancel_dblstrike_font();
              break;

    case 'M': set_char_width(CPI12);
              break;

    case 'N': /* Set Skip Perforation - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    case 'O': /* Cancel Skip Perforation - ** IGNORE ** */
              break;

    case 'P': set_char_width(CPI10);
              break;

    case 'Q': /* Set Right Margin */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              frm.r_margin = (float) *param * frm.char_width;
              free(param);
              break;

    case 'R': /* Select International Character Set - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    case 'S': /* Set sub/super script */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              set_sub_super_script(*param);
              free(param);
              break;

    case 'T': /* Cancel sub/super script */
              cancel_sub_super_script();
              break;

    case 'U':   /* Select or Cancel Uni-directional printing - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    case 'W':   /* Select or Cancel Double-width - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    case 'g': set_char_width(CPI15);
              break;

    case 'i':   /* Select or Cancel incremental print - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    case 'l':  /* Set Left Margin */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              frm.l_margin = (float) *param * frm.char_width;
              free(param);
              break;

    case 'm':   /* Select Character Table - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    case 'r':   /* Select Colour */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              set_colour_value(*param);
              free(param);
              break;

    case 's':   /* Select or Cancel half speed - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    case 'p':  /* Proportional font - treat as NLQ */
    case 'x':  /* Set or Cancel NLQ */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              if(param)
                set_nlq_font();
              else
                cancel_nlq_font();
              free(param);
              break;

    case '\x19':  /* Select Paper Bin - ** IGNORE ** */
              param = malloc(sizeof(int));
              read_parameters(param, 1, 1);
              free(param);
              break;

    /* Silently ignore these codes */
    case '\x0E':  /* Select Double Width */
    case '\x23':  /* Accept 8th Bit */
              break;

    default:
              esc_error(action);
              break;
  }
}

int process()
{
int x;

  strcpy(sts.curr_font, cfg.std_font); 

  /* Set ready for top line */
  if(start_page())
    return -1;

  /* Process 'text file' to output */
  while((x = fgetc(ifile)) != EOF)
  {
    if(sts.cur_col >= frm.r_margin)
      ctrl_code('\x0a'); /* Send a line feed command */

    switch(x)
    {
      case '\x08':  /* Back Space */
      case '\x09':  /* Horizontal Tab */
      case '\x0a':  /* Line Feed */
      case '\x0b':  /* Vertical Tab */
      case '\x0c':  /* Form Feed */
      case '\x0d':  /* Carriage Return */
      case '\x0f':  /* Select Condensed */
      case '\x12':  /* Cancel Condensed */
        ctrl_code(x);
        break;

      case '\x1B':  /* ESC code */
        escape(fgetc(ifile));
        break;

      /* Silently ignore these codes */
      case '\x07':  /* Sound alarm */
      case '\x0e':  /* Select double width */
      case '\x11':  /* Device control 1 */
      case '\x13':  /* Device control 3 */
      case '\x14':  /* Cancel double width */
      case '\x18':  /* Clear buffer */
      case '\x7f':  /* Delete last character from buffer */
        break;

      default:
        if(isprint(x))
        {
          if(sts.new_page == FF_ON)
            start_page();

          if(sts.font_set == FF_OFF)
            set_new_font();

          if(sts.text_mode == FF_OFF)
          {
            fprintf(ofile,"(");
            sts.start_col = sts.cur_col;
            sts.text_mode = FF_ON;
          }
          if(x == '\\' || x == '(' || x == ')')
            fputc('\\', ofile);

          fputc(x, ofile);

          if(sts.underline && (sts.ul_start == 0) && (sts.ul_end == 0))
            sts.ul_start = sts.cur_col;

          sts.cur_col += frm.char_width;

          if(sts.underline)
            sts.ul_end = sts.cur_col;
        }
        break;
    }

    /* Jump to next page if current page is full */
    if(sts.cur_line <= 0)
      ctrl_code('\x0c');
  }

  if(sts.text_mode == FF_ON)
  {
    fprintf(ofile,") %0.2f %0.2f ms\n", sts.start_col, sts.cur_line);
    sts.text_mode = FF_OFF;
  }

  if(sts.new_page == FF_OFF)
    fprintf(ofile,"showpage\n\n");

  write_file_end();

return 0;
}

int main(int argc, char *argv[])
{
char form_file[FN_LEN + 1];
int opt, cfg_flg = 0, frm_flg = 0, err_flg = 0;
time_t tim;
struct tm *tm;
int i, run_mode = 0;

  strcpy(config_file, _CFG_FILE);
  form_file[0] = 0;

  ifile = stdin;
  ofile = stdout;
  verbose = 0;

  while((opt = getopt(argc, argv, "c:f:hi:lo:tv")) != -1)
  {
    switch(opt)
    {
      case 'c': if(!cfg_flg)
                {
                  if(strlen(optarg) >= FULLPATH)
                    err_flg = 1;
                  else
                  {
                    strcpy(config_file, optarg);
                    cfg_flg = 1;
                  }
                }
                else
                  err_flg = 2;
                break;

      case 'f': if(!frm_flg)
                {
                  if(strlen(optarg) >= FN_LEN)
                    err_flg = 3;
                  else
                  {
                    strcpy(form_file, optarg);
                    frm_flg = 1;
                  }
                }
                else
                  err_flg = 4;
                break;

      case 'i': ifile = fopen(optarg, "r");
                if(!ifile)
                {
                  fprintf(stderr, "%s: %s: %s\n", argv[0], optarg, 
                          strerror(errno));
                  exit(EXIT_FAILURE);
                }
                break;

      case 'o': ofile = fopen(optarg, "w");
                if(!ofile)
                {
                  fprintf(stderr, "%s: %s: %s\n", argv[0], optarg, 
                          strerror(errno));
                  exit(EXIT_FAILURE);
                }
                break;

      case 't': run_mode = 1;
                break;

      case 'h':
      case 'v': fprintf(stderr, "%s  %s\n%s\n\n", PROGNAME, VERSION, COPYRIGHT);
                exit(EXIT_SUCCESS);
                break;

      case 'l':	verbose;
                break;

      default:  err_flg = 99;
                break;

    }
  }

  if(err_flg)
  {
    syslog(LOG_ERR, "Invalid command line parameter - %d", err_flg);
    exit(EXIT_FAILURE);
  }

  err_flg = get_config(config_file, run_mode);

  if(run_mode)
    exit(EXIT_SUCCESS);

  if(err_flg)
    exit(EXIT_FAILURE);
  else
  {
    /* Create date string. */
    time(&tim);
    tm = (struct tm *)localtime(&tim);

    date_string = strdup(asctime(tm));
    i = strlen (date_string);
    date_string[i - 1] = '\0';


    /* Get user's passwd entry. */
    pswd = getpwuid (getuid ());
    if (pswd == NULL)
    {
      syslog(LOG_ERR, "Couldn't find passwd entry for UID - %d", getuid());
      exit(EXIT_FAILURE);
    }

    /* Initialize the "printer" */
    init();

    /* Set up for default form */
    sts.curr_form[0] = 0;
    if(form_file[0])
    {
      strcpy(sts.curr_form, cfg.forms_dir);
      strcat(sts.curr_form, "/");
      strcat(sts.curr_form, form_file);
    }
    else
      syslog(LOG_INFO, "No form file specified");

    sts.new_page = FF_OFF;
    process();
  }

exit(EXIT_SUCCESS);
}

/***********************************************************************/
/*                                                                     */
/*                          End of Source Code                         */
/*                                                                     */
/***********************************************************************/

