/***********************************************************************/
/*                                                                     */
/*  FenForm - Document Printer Header File                             */
/*  Vn 1.4                                                             */
/*  Author - Mike Eggleston                                            */
/*  Copyright 2005 Fenland Software Ltd                                */
/*  Date 05/01/2005                                                    */
/*                                                                     */
/***********************************************************************/

#ifndef FENFORM_H
#define FENFORM_H

#include <ctype.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>

// Paper Sizes
#define A4_LENGTH	842
#define A4_WIDTH	595

#define DFT_LPI		6
#define DFT_CPI		10
#define DFT_PAPER_L	A4_LENGTH
#define DFT_PAPER_W	A4_WIDTH

#define LPI6		12
#define LPI8		9
#define LPI72		1
#define LPI7_72		7

#define CPI10		7.2
#define CPI12		6.0
#define CPI15		4.8
#define CPI17		4.21
#define CPI18		4.0
#define CPI20		3.6

#define FF_ON		1
#define FF_OFF		0
#define DFT_HTAB	8
#define SUBSCRIPT	1
#define SUPERSCRIPT	2


/* Default field (array) sizes           */
#define BUFFER_LEN	1024
#define LINE_LN		80
#define FN_LEN		64
#define FULLPATH	256
#define EMAILLENGTH	4096
#define MAX_V_TABS	16
#define MAX_H_TABS	32
#define BASE_LINEWIDTH	0.7

/* Configuration Defaults                */
#define _CFG_FILE		"/etc/fenform.cfg"
#define _FORMS			"/var/spool/forms"
#define _STDFONT		"Courier"
#define _STDBOLD		"Courier-Bold"
#define _STDDBLSTRIKE		"Courier-Bold"
#define _STDITALIC		"Courier-Oblique"
#define _STDBOLDITALIC		"Courier-BoldOblique"
#define _STDDBLSTRIKEITALIC	"Courier-BoldOblique"
#define _NLQFONT		"Times-Roman"
#define _NLQBOLD		"Times-Bold"
#define _NLQDBLSTRIKE		"Times-Bold"
#define _NLQITALIC		"Times-Italic"
#define _NLQBOLDITALIC		"Times-BoldItalic"
#define _NLQDBLSTRIKEITALIC	"Times-BoldItalic"
#define _HOFFSET		0
#define _VOFFSET		0
#define _MAILER			"/usr/bin/sendmail"
#define _COLOUR_RED		0xFF0000
#define _COLOUR_GREEN		0x00AA00
#define _COLOUR_BLUE		0x0000FF
#define _COLOUR_YELLOW		0xFFFF00
#define _COLOUR_VIOLET		0xFF00FF
#define _COLOUR_ORANGE		0xFF5500

/* Fenform configuration parameters      */
struct FF_config
{
char forms_dir[FULLPATH + 1];
char std_font[FN_LEN + 1];
char bold_font[FN_LEN + 1];
char dblstrike_font[FN_LEN + 1];
char italic_font[FN_LEN + 1];
char bold_italic_font[FN_LEN + 1];
char dblstrike_italic_font[FN_LEN + 1];
char nlq_font[FN_LEN + 1];
char bold_nlq_font[FN_LEN + 1];
char dblstrike_nlq_font[FN_LEN + 1];
char nlq_italic_font[FN_LEN + 1];
char bold_nlq_italic_font[FN_LEN + 1];
char dblstrike_nlq_italic_font[FN_LEN + 1];
float line_width;
float h_offset;
float v_offset;
unsigned int full_colour;
char mailer[FULLPATH + 1];
char admin[FULLPATH + 1];
};

struct FF_layout
{
char form_name[FN_LEN + 1];
char orientation;
float form_length;
float form_width;
float line_space;
float char_width;
float l_margin;
float r_margin;
float htabs[MAX_H_TABS];
float vtabs[MAX_V_TABS];
};

struct FF_status
{
char curr_form[FULLPATH + 1];
char curr_font[FN_LEN + 1];
float red;
float green;
float blue;
unsigned int font_set;
unsigned int page;
float cur_line;
float start_col;
float cur_col;
unsigned int h_tab;
unsigned int v_tab;
unsigned int italics;
float italics_start;
float italics_end;
unsigned int bold;
float bold_start;
float bold_end;
unsigned int dblstrike;
float dblstrike_start;
float dblstrike_end;
unsigned int nlq;
float nlq_start;
float nlq_end;
unsigned int underline;
float ul_start;
float ul_end;
unsigned int sub_superscript;
unsigned int text_mode;
float prev_char_width;
float line_offset;
unsigned int new_page;
};

struct FF_barcode
{
unsigned int type;
unsigned int flags;
unsigned int height;
unsigned int width;
};

#endif  /* If Not FENFORM_H */

