/***********************************************************************/
/*                                                                     */
/*  PS2FORM - Convert a PostScript file to a form                      */
/*  Vn 1.0                                                             */
/*  Author - Mike Eggleston                                            */
/*  Copyright 2005 Fenland Software Ltd                                */
/*  Date 25/01/2005                                                    */
/*                                                                     */
/***********************************************************************/

#define  PROGNAME	"PS2FORM"
#define  VERSION	"Vn 1.0"
#define  COPYRIGHT	"Copyright 2005 Fenland Software Ltd"

#include "fenform.h"

/******************************/
/*  Global Variables          */
/******************************/
void show_help(void)
{
  printf("Usage: ps2form -i input-file -f form-file\n");
  printf("       ps2form -v\tShow version number and exit\n");
  printf("       ps2form -h\tThis help page\n");
}

int find_last(char *s)
{
int n;
char *substr;

  if(strncasecmp(s, "showpage", 8) == '\0')
    return 1;

  substr = strstr(s, "showpage");
  if(substr)
  {
    *substr = '\0';
    return 1;
  }

return 0;
}

int find_pagesize(FILE *fp, int *l, int *w)
{
char s[BUFFER_LEN + 1], *substr;
int rtn = 0;

  if(fp)
  {
    do
    {
      if(!(fgets(s, BUFFER_LEN, fp)))
        break;
    }
    while (strncasecmp(s, "%%BeginPageSetup", 16));

    do
    {
      if(!(fgets(s, BUFFER_LEN, fp)))
        break;

      if(strstr(s, "setpagesize"))
      {
        *w = atoi(s);
        substr = strchr(s, ' ');
        *l = atoi(substr);
        rtn = 1;
        break;
      }
    }
    while (s);
  }

return(rtn);
}
  
int process(char *in_file, char *out_file, int lpi, int cpi, int voff, int hoff)
{
FILE *FF_ps, *FF_frm;
int prolog = 0;
int page_length, page_width;
char str[BUFFER_LEN + 1];
char form_name[FN_LEN + 1];
char *ptr;

  /* Open input file */
  FF_ps = 0;
  fprintf(stderr, "Opening  %s\n",in_file);
  if(in_file[0])
  {
    if(!(FF_ps = fopen(in_file, "r")))
    {
      fprintf(stderr, "OPEN failed - %s  %s\n",in_file, strerror(errno));
      return 1;  /* Exit if input file not found. */
    }
  }

  /* Open output file */
  FF_frm = 0;
  if(out_file[0])
  {
    if(!(FF_frm = fopen(out_file, "w")))
    {
      fprintf(stderr, "OPEN failed - %s  %s\n",out_file, strerror(errno));
      return 1;  /* Exit if output file not found. */
    }
  }

  ptr = strrchr(out_file, '/');
  if (ptr)
    strcpy(form_name, ptr + 1);
  else
    strcpy(form_name, out_file);
  
  if(FF_ps)
  {
    if(!find_pagesize(FF_ps, &page_length, &page_width))
    {
      page_length = DFT_PAPER_L;
      page_width = DFT_PAPER_W;
    }
    fseek(FF_ps, 0L, SEEK_SET);

    if(fgets(str, BUFFER_LEN, FF_ps))
    {
      if(strncmp(str, "%!PS", 4))
      {
        fprintf(stderr, "ERROR - Unrecognized input file.\n");
        return 2;
      }
    }
    else
      return 1;

    fprintf(FF_frm, "*PARMS\n");
    fprintf(FF_frm, "NAME=%s\n", form_name);
    fprintf(FF_frm, "LPI=%d\n", lpi);
    fprintf(FF_frm, "CPI=%d\n", cpi);
    fprintf(FF_frm, "LENGTH=%d\nWIDTH=%d\n", page_length, page_width);
    if(voff != 0)
      fprintf(FF_frm, "VOFFSET=%d\n", voff);
    if(hoff != 0)
      fprintf(FF_frm, "HOFFSET=%d\n", hoff);
    fprintf(FF_frm, "**END\n\n");

    while(fgets(str, BUFFER_LEN, FF_ps))
    {
      if(strncasecmp(str, "%%BeginProlog", 13) == '\0')
        prolog = 1;

      if(find_last(str))
      {
        fputs(str, FF_frm);
        break;
      }

      if(strncmp(str, "%%", 2) || prolog)
        fputs(str, FF_frm);

      if(strncasecmp(str, "%%EndProlog", 11) == '\0')
        prolog = 0;
    }
  }

  fclose(FF_ps);
  fclose(FF_frm);

return 0;
}


int main(int argc, char *argv[])
{
char *ps_file;
char *form_file;

int opt, ps_flg = 0, frm_flg = 0, err_flg = 0;
int n, lpi = DFT_LPI, cpi = DFT_CPI, voff = 0, hoff = 0;

  while((opt = getopt(argc, argv, "c:f:i:l:hv")) != -1)
  {
    switch(opt)
    {
      case 'i': if(!ps_flg)
                {
                  if(ps_file = malloc(strlen(optarg) + 1))
                  {
                    strcpy(ps_file, optarg);
                    ps_flg = 1;
                    if(!frm_flg)
                    {
                      if(form_file = malloc(strlen(optarg) + 6))
                      {
                        if(strrchr(ps_file, '/'))
                          strcpy(form_file, strrchr(ps_file, '/') + 1);
                        else
                          strcpy(form_file, ps_file);

                        for (n=strlen(form_file); n; n--)
                        {
                          if(form_file[n] == '.')
                          {
                            form_file[n] = 0;
                            break;
                          }
                        }
                        strcat(form_file, ".form");
                      }
                    }
                  }
                  else
                    err_flg = 2;
                }
                else
                  err_flg = 3;
                break;

      case 'f': if(!frm_flg)
                {
                  if(form_file = realloc(form_file, strlen(optarg) + 5))
                  {
                    strcpy(form_file, optarg);
                    frm_flg = 1;
                  }
                  else
                    err_flg = 5;
                }
                else
                  err_flg = 6;
                break;

      case 'c': cpi = atoi(optarg);
                break;

      case 'l': lpi = atoi(optarg);
                break;

      case 'v': printf("%s  %s\n%s\n\n", PROGNAME, VERSION, COPYRIGHT);
                exit(EXIT_SUCCESS);
                break;

      case 'h':
      default:  printf("%s %s\n%s\n", PROGNAME, VERSION, COPYRIGHT);
                show_help();
                exit(EXIT_SUCCESS);
                break;

    }
  }

  if (!err_flg && !form_file[0])
    err_flg = 4;
 
  if (!err_flg && !ps_flg)
    err_flg = 1;
 
  if (err_flg)
  {
    printf("%s %s\n%s\n", PROGNAME, VERSION, COPYRIGHT);
    switch (err_flg)
    {
      case 1: printf("ERROR - No input file specified.\n");
              break;

      case 2: printf("ERROR - Unable to allocate memory for input file.\n");
              break;

      case 3: printf("ERROR - Input file specified more than once.\n");
              break;

      case 4: printf("ERROR - No form file specified.\n");
              break;

      case 5: printf("ERROR - Unable to allocate memory for form file.\n");
              break;

      case 6: printf("ERROR - Form file specified more than once.\n");
              break;

    }
    show_help();
    exit(EXIT_FAILURE);
  }
  else
    process(ps_file, form_file, lpi, cpi, voff, hoff);

exit(EXIT_SUCCESS);
}

/***********************************************************************/
/*                                                                     */
/*                          End of Source Code                         */
/*                                                                     */
/***********************************************************************/

